"""
xorisso functions for dmm (not yet implemented)
"""

import os

def recipe_run(config, globalconf):
    """
    xorrisso actions:

     - geniso generates an iso image
    """
    if config['action'] == "geniso":
        generate_iso(config["isopath"], config["live-media-path"])


def generate_iso(isopath, live_media_path):
    
    """
    Generate an ISO image using xorriso
    """
    print("xorriso -outdev %s -volid d-live testing ci amd64 -padding 0 -compliance no_emul_toc -map %s -chmod 0755 / -- -boot_image isolinux dir=/isolinux -boot_image isolinux system_area=/usr/lib/grub/i386-pc/boot_hybrid.img -boot_image any next -boot_image any efi_path=boot/grub/efi.img -boot_image isolinux partition_entry=gpt_basdat % (isopath, live-media-path)")




"""
reprepro functions for for dmm.
"""

import os
import pathlib

def recipe_run(config, globalconf):
    """
    Perform actions for reprepro module
    """
    #TODO: break all these into smaller functions outside this one
    pkgpool_path = globalconf['workspace'] + "/pkgpool"
    path = pathlib.Path(globalconf['workspace'] + "/pkgpool/conf")
    path.mkdir(parents=True, exist_ok=True)

    package_path = config['prepend_path'] + "/" + config['package_path']
    version = globalconf['release']
    architectures = globalconf['architecture']
    components = config['components']
    description = globalconf['description']
    name = globalconf['name']

    reprepro_config = ("Origin: %s\nLabel: %s\nSuite: %s\nVersion: %s\nDescription: %s\n"
                       "Codename: %s\nComponents: %s\nArchitectures: %s\nUDebComponents: %s\n" %
                       (name, name, version, version, description, version, components, architectures, components))
    text_file = open(pkgpool_path + "/conf/distributions", "w")
    sources_write = text_file.write(reprepro_config)
    text_file.close()

    path = pathlib.Path(config['live-media-path'])
    path.mkdir(parents=True, exist_ok=True)

    # generate package pool
    # - usual debs:
    os.system('reprepro --ignore=extension -b %s includedeb %s %s' % (pkgpool_path, version, package_path + "/*.deb"))
    # - udebs, for installer:
    os.system('reprepro --ignore=extension -b %s includeudeb %s %s' % (pkgpool_path, version, package_path + "/*.udeb"))

    # Copy pkgpool to media
    os.system('cp -r %s %s %s' % (pkgpool_path + "/dists", pkgpool_path + "/pool", config['live-media-path'] + "/"))

    if config['clean_package_path']:
        os.system('rm %s/*.deb' % (config['prepend_path'] + config['package_path']))




"""
Interact with systemd tools and services.
"""

import subprocess
import json

def list_timezones():
    """
    Just return a list of available timezones
    """
    timezones = subprocess.check_output(["/usr/bin/timezonectl", "list-timezones",
                                            ]).decode("utf-8").split("\n")
    return timezones




"""
bootstrapping debian functions for for dmm.
"""

from command_runner import command_runner

def debootstrap(destination, release, mirror, debootstrapopts):
    """
    Use debootstrap to bootstrap a Debian system.
    """
    ecode, result = command_runner("debootstrap %s %s %s %s" % (debootstrapopts,
                                                    release,
                                                    destination,
                                                    mirror))
    return ecode, result


def mmdebstrap(destination, release, mirror, debootstrapopts):
    """
    Use mmdebstrap to bootstrap a Debian system.
    """
    print("mmdebstrap %s %s" % (release, destination))
    ecode, result = command_runner("mmdebstrap %s %s" % (release, destination))
    return ecode, result




"""
vmsetup functions for for dmm.
"""

import os
from command_runner import command_runner

def vmsetup(chroot, root_during_build, root_dev_in_vm):
    """
    Additional basic setup for a virtual machine.
    """
    os.system("sed -i 's/%s/%s/g' %s/boot/grub/grub.cfg" % (root_during_build,
                                                            root_dev_in_vm,
                                                            chroot))
    os.system("echo '/dev/%s / ext4 defaults 0 0' > %s/etc/fstab" % (root_dev_in_vm, chroot))

    # qemu's bios only reads the EFI files if it can find them in the right place, and
    # vfat doesn't support symlinks
    os.system("cp -r %s/boot/efi/EFI/debian %s/boot/efi/EFI/boot" % (chroot, chroot))
    os.system("cp -r %s/boot/efi/EFI/boot/grubx64.efi %s/boot/efi/EFI/boot/bootx64.efi" % (chroot, chroot))
    os.system(("sed -i 's/quiet/quiet console=ttyS0/g' %s/etc/default/grub") % chroot)
    os.system("chroot %s /usr/sbin/update-grub" % chroot)





"""
networking functions for for dmm.
"""

def set_chroot_hostname(chroot, hostname):
    """
    Writes a hostname in a chroot
    Could be replaced by writefile?
    """
    text_file = open(chroot + "/etc/hostname", "w")
    hostname_write = text_file.write(hostname)
    text_file.close()




"""
DMM Recipe runner
"""

# Imports
import importlib
import sys
import yaml
from loguru import logger

# Initialize logger
logger.add(sys.stderr, format="{time} {level} {message}",
           filter="my_module", level="INFO")
logger.add("out.log", backtrace=True, diagnose=True)


@logger.catch
def read_recipe(recipe):
    """
    Read the list of stuff that we should be doing.
    """
    logger.info(f"Reading recipe: {recipe}")
    configfile = open(recipe, "r")
    global config
    config = (yaml.safe_load(configfile))
    logger.debug("Config:\n" + str(config))
    sys.path.append(config["module_path"])
    logger.debug("Add-on module path(s): ", config['module_path'])
    return list(config['recipe'])


@logger.catch
def perform_recipe(recipe):
    """
    Runs through the tasks in our recipe.
    """
    for task in read_recipe(recipe):
        run_task(task)
    logger.info("Recipe: %s has completed." % recipe)


@logger.catch
def run_task(task):
    """
    Run through a task in a recipe.

    We load a module as needed and then unload it when we're done
    to conserve memory for installer environments.
    """
    logger.info(f"Running task: {task}")
    module_name = config['recipe'][task]["module"]
    task_config = config['recipe'][task]

    try:
        module = importlib.import_module(f"dmm.{module_name}")
    except ModuleNotFoundError:
        logger.error(f"Module: {module_name} could not be found")

    try:
        function = getattr(module, config["recipe"][task]["function"])
        task_params = {key: value for key, value in task_config.items()
                       if key not in ["module", "function"]}
        ecode, result = function(**task_params)
        logger.debug(f"Error code : {ecode}, result : {result}")
    except AttributeError:
        logger.error(f"Module: {module_name} doesn't seem to contain the "
                     f"function: {config['recipe'][task]["function"]}")


logger.debug("DMM Recipe reader loaded")



"""
timezone functions for for dmm.
"""

from command_runner import command_runner

def list_timezones():
    """
    Use systemd to list all timezones.
    """
    ecode, result = command_runner("timedatectl list-timezones")
    return (ecode, result)


def set_timezone(timezone):
    """
    Set a timezone
    """
    print("doesn't actually do anything yet")




"""
disk and partition functions for dmm.
"""

import pathlib
from command_runner import command_runner

def run_parted_commands(disk, command):
    """
    Iterate through parted commands and return an error if applicable.
    """
    ecode, result = command_runner("parted -s -- %s %s" % (disk, command))
    return ecode, result


def run_parted_command_set(command_set):
    """
    Iterate through a set of parted commands and return an error if applicable.
    """
    #print(command_set)
    for partition in command_set:
        #print("")
        #print("Command: ", partition)
        #print("action-type: ", partition['action-type'])
        #print("command: ", partition['action-type'])
        #print(repr(("device: ", partition['device'])))
        #print(partition['device'], partition['command'])
        ecode, result = run_parted_commands(partition['device'], partition['command'])
        ecode, result =  command_runner("mkfs.%s %s" % (partition['fstype'], partition['partition'])) 
        if partition["espflag"]:
            # FIXME: Unhardcode
            ecode, result = command_runner("parted -s -- set 1 esp on")
        print (ecode, result)
    return ecode, result 


def initialize_gpt_table(disk):
    """
    Create a new GPT partition table on the device.
    """
    command_runner("wipefs -a %s" % (disk))
    ecode, result = command_runner("sgdisk -o %s" % (disk))
    return ecode, result


def losetup(diskimg, mountpoint, loopdev):
    """
    Sets up a loopback device.
    """
    ecode, result = command_runner("losetup -P /dev/%s %s" % (loopdev, diskimg))
    return ecode, result


def losetup_teardown(loopdev):
    """
    Unmounts a loopback device.
    """
    ecode, result = command_runner("losetup -d /dev/%s" % (loopdev))
    return ecode, result


def create_filesystems(partitions):
    """
    Creates a new filesystem on a device.
    """
    for partition in partitions:
        ecode, result = command_runner("mkfs.%s %s %s" % (partition['fstype'], partition['options'], partition['partition']))
        if ecode != 0:
            return ecode, result


def mount(filesystems):
    """
    Mounts filesystems
    """
    for filesystem in filesystems:
        path = pathlib.Path(filesystem['mountpoint'])
        path.mkdir(parents=True, exist_ok=True)
        ecode, result = command_runner("mount -t %s %s %s %s" % (filesystem['fstype'], filesystem['mountopts'],
                  filesystem['source'], filesystem['mountpoint']))
        if ecode != 0:
            return ecode, result
    return ecode, result


def umount(mounts):
    """
    Unmounts filesystems
    """
    for mount in mounts:
        path = pathlib.Path(mount['mountpoint'])
        ecode, result = command_runner("umount %s" % (mount['mountpoint']))
        if ecode != 0:
            return ecode, result
    return ecode, result



"""
grub functions for for dmm.
"""

import os
from command_runner import command_runner

def grub_install(chroot, platform):
    """
    Install an appropriate grub version
    """
    ecode, result = command_runner("chroot %s apt-get -y -qq install grub-%s" % (chroot, platform))
    ecode, result = command_runner("chroot %s update-grub" % chroot)
    if platform == "efi":
        ecode, result = command_runner("chroot %s grub-install --force-extra-removable --target=x86_64-efi" % chroot)
    # hmm what happens if we have biosboot?
    return ecode, result


def recipe_run(config, globalconf):
    """
    Perform actions for grub module
    """
    if config['action'] == "mkrescue":
        return os.system('grub-mkrescue --modules="multiboot fat linux part_msdos minicmd ls iso9660" --output="%s" %s' % (config['isopath'], globalconf["live-media-path"]))
    if config['action'] == "create_legacy_boot_img":
        setup_grub_legacy_boot(globalconf["live-media-path"])


def setup_grub_legacy_boot(path):
    """
    Create PC BIOS (amd64/i386 legacy) bootable grub image.
    """
    # TODO: should probably happen inside the chroot?
    os.system('grub-mkstandalone --directory=/usr/lib/grub/i386-pc --format=i386-pc '
              '--themes="" --fonts="" --locales="" --modules="linux normal iso9660 '
              'biosdisk search png gfxmenu" --install-modules="linux normal iso9660 '
              'biosdisk memdisk search tar ls png gfxmenu" --output core.img')
    os.system('cat /usr/lib/grub/i386-pc/cdboot.img core.img > bios.img')


def depends():
    """
    Returns a list of dependencies for this module.
    Currently this is Debian packages.
    """
    return ({'xorriso': {"priority": "required", \
             "description": "command line ISO-9660 and Rock Ridge manipulation tool"},
             'mtools': {"priority": "required", \
             "description": "Tools for manipulating MSDOS files"}
           })




"""
create users, currently only in chroot
"""

import os
from command_runner import command_runner

def add_users(users, chroot):
    """
    Add one or more users to the system.
    """
    for user in users:
        password_hash = os.popen("mkpasswd --method=SHA-512 " + user['password']).read().replace('\n', '')
        ecode, result = command_runner("chroot %s useradd -p '%s' %s" % (chroot, password_hash, user['username']))
        ecode, result = command_runner("chroot %s mkdir home/%s" % (chroot, user['username']))
        ecode, result = command_runner("chroot %s chown %s home/%s" % (chroot, user['username'], user['username']))
        if user['sudo']:
            ecode, result = command_runner("chroot %s adduser %s sudo" % (chroot, user['username']))
            ecode, result = command_runner("chroot %s apt-get -y install sudo" % chroot)
    return ecode, result




"""
debian functions for for dmm.
"""

import subprocess
from command_runner import command_runner

def update_sources(sources_list, sources_file, update_sources, chroot):
    """
    Update sources.list and optionally perform an apt-get update.
    """
    text_file = open(chroot + sources_file, "w")
    sources_write = text_file.write(sources_list)
    text_file.close()
    if update_sources:
        ecode, result = command_runner("chroot %s apt-get -q update" % (chroot),
                                shell=True)
        return ecode, result




"""
apt functions for for dmm.
"""

from command_runner import command_runner


def install(packages="", chroot=None, **kwargs):
    """
    Install APT packages.

    chroot - specify chroot to do this in, False for host.
    """
    if chroot:
        ecode, result = command_runner('LANG=C LANGUAGE=C LC_TYPE=C '
                                       'LC_MESSAGES=C '
                                       'LC_ALL=C DEBIAN_FRONTEND=noninteractive '
                                       'chroot %s apt-get -y install %s'
                                       % (chroot,  packages), shell=True)
    else:
        ecode, result = command_runner('LANG=C LANGUAGE=C LC_TYPE=C '
                                       'LC_MESSAGES=C LC_ALL=C '
                                       'DEBIAN_FRONTEND=noninteractive '
                                       'apt-get -y install %s' %
                                       (packages), shell=True,
                                        live_output=True)
    return ecode, result


def clean(chroot=None):
    """
    Clean apt archives.

    chroot - specify chroot to do this in, False for host.
    """
    if chroot:
        ecode, result = command_runner('chroot %s apt-get clean' % chroot)
    else:
        ecode, result = command_runner('apt-get clean')
    return ecode, result


def update(chroot=None):
    """
    Update APT archives.
    """
    if chroot:
        ecode, result = command_runner('chroot %s apt-get -q update'
                                       % (chroot), shell=True,
                                          live_output=True)
    else:
        ecode, result = command_runner('apt-get -q update',
                                       shell=True, live_output=True)
    return ecode, result


def download_only(chroot=None, packages="", **kwargs):
    """
    Download packages for apt cache.
    For download-only, we do this one package at a time, since it's possible
    that some packages might conflict, and you might want to ship conflicting
    files on media.

    chroot - specify chroot to do this in, False for host.
    """
    for package in packages.split():
        if chroot:
            ecode, result = command_runner('chroot %s apt-get -d -y --reinstall install %s'
                      % (chroot, package))
        else:
            ecode, result = command_runner('apt-get -d -y --reinstall install %s' % (package))
    return ecode, result




#!/usr/bin/python3
from dmm.modules.fallocate import fallocate
import os

def test_fallocate_create_file():
    fallocate.create_file("4M", "/tmp/test.img")
    file_stats = os.stat("/tmp/test.img")
    # 4M is 4194304
    assert file_stats.st_size == 4194304
    os.remove("/tmp/test.img")



"""
Show disk / partition information.
"""

import subprocess
import json

def list_scsi_devices():
    """
    Just return a list of available scsi disks.
    """
    lsblk_output = subprocess.check_output(["/bin/lsblk", "--scsi",
                                            "-noheadings", "--output",
                                            "NAME", "-n"]).decode("utf-8").split("\n")
    scsi_devices = [x for x in lsblk_output if x]
    return scsi_devices


def list_block_devices():
    """
    Get json data from lsbkl and convert it to a python dict.
    """
    lsblk_output = subprocess.check_output(["/bin/lsblk", "-J", "-O", "-b"])
    json_data = json.loads(lsblk_output)
    print(json_data)
    return(json_data)


def get_device_information(device):
    """
    Get ..."
    """


def list_partitions():
    """
    TODO: We need to list the partitions somehow
    so that we can present a disk to run cfdisk on.
    """



#!/usr/bin/python3
from dmm.modules.dd import dd
import os

def test_dd_create_file():
    dd.create_file("/dev/zero", "/tmp/test.img", "2M", "2", 2)
    file_stats = os.stat("/tmp/test.img")
    # 2M * 2 should equal 4M
    assert file_stats.st_size == 4194304
    os.remove("/tmp/test.img")



""" Wrapper around fdisk """
import os


from subprocess import (PIPE, Popen)


def invoke(command):
    '''
    Invoke process and return its output.
    '''
    return Popen(command, stdout=PIPE, shell=True).stdout.read()


def list_disks():
    disklist = invoke('fdisk -l')
    print(disklist)



"""
download remote content for dmm
"""

import os
import pathlib

def init():
    """
    Initialization for download module
    """


def recipe_run(config, globalconf):
    """
    Perform actions for download module
    """
    if config['action'] == 'download_files':
        for download in config['files']:
            download_file(download['url'], download['destination'])


def download_file(url, destination):
    """
    Download file using curl and store it at location
    """
    print("Downloading %s to %s" % (url, destination))
    # Ensure that base path exists
    path = pathlib.Path(destination).parent
    path.mkdir(parents=True, exist_ok=True)
    os.system("curl %s -o %s" % (url, destination))


init()



"""
livedisk functions for for dmm.
"""

import pathlib
import os
from shutil import copyfile

def init():
    """
    Initialization for livedisk module
    """


def recipe_run(config, globalconf):
    """
    Perform actions for livedisk module
    """
    # Create working directory for live media
    path = pathlib.Path(globalconf['live-media-path'] + "/live")
    path.mkdir(parents=True, exist_ok=True)

    # copy squashfs, kernel and initramfs into livedir
    print("Current working directory is " + os.getcwd())
    squashfs = globalconf['squashfs']
    kernel = globalconf['chroot'] +  "/boot/vmlinuz*"
    initrd = globalconf['chroot'] + "/boot/initrd*"
    livepath = globalconf['live-media-path'] + "/live/"
    os.system("cp %s %s" % (squashfs, livepath + "/filesystem.squashfs"))
    os.system("cp %s %s" % (kernel, livepath + "/vmlinuz"))
    os.system("cp %s %s" % (initrd, livepath + "/initrd.gz"))

    # Copy template directory to live media
    teamplate_dir = config['template_dir']
    os.system("cp -r %s/ %s" % (config['template_dir'] + "/*", globalconf['live-media-path']))

    # Update initramfs
    if config['update-initramfs']:
        os.system("chroot %s update-initramfs -u" % globalconf['chroot'])

    # TODO: create md5sums



init()



"""
Python wrapper for dd, usually to create large empty disk image
"""

import subprocess

def create_file(ifile, ofile, bs, count, timeout):
    """
    Create or overwrite a file using dd.

    As with dd:

    - ifile: inputfile
    - ofile: outputfile
    - bs: size of block
    - count: how many blocks should be written
    - timeout: make the operation timeout after amount of seconds
    """
    dd_output = subprocess.run(["/usr/bin/dd",
                        "if=" + ifile,
                        "of=" + ofile,
                        "bs=" +  bs,
                        "count=" + count],
                        check=True,
                        timeout=timeout,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        universal_newlines=True,)
    return (dd_output.returncode, dd_output.stdout, dd_output.stderr)




"""
sbuild functions for for dmm.
"""

import os
import subprocess

def install(chroot, packages):
    """
    Install APT packages.

    chroot - specify chroot to do this in, False for host.
    """
    if chroot:
        result = os.system('LANG=C LANGUAGE=C LC_TYPE=C LC_MESSAGES=C '
                           'LC_ALL=C DEBIAN_FRONTEND=noninteractive '
                           'chroot %s apt-get -y install %s'
                           % (chroot,  packages))
        return result
    else:
        result = os.system('LANG=C LANGUAGE=C LC_TYPE=C LC_MESSAGES=C '
                           'LC_ALL=C DEBIAN_FRONTEND=noninteractive '
                           'apt-get -y install %s' % (packages))
        return result


def clean(chroot):
    """
    Clean apt archives.

    chroot - specify chroot to do this in, False for host.
    """
    if chroot:
        os.system('chroot %s apt-get clean' % chroot)
    else:
        os.system('apt-get clean')


def update(chroot):
    """
    Update APT archives.
    """
    if chroot:
        os.system('chroot %s apt-get -q update' % (chroot))
    else:
        os.system('apt-get -q update')


def download_only(chroot, packages):
    """
    Download packages for apt cache.
    For download-only, we do this one package at a time, since it's possible
    that some packages might conflict, and you might want to ship conflicting
    files on media.

    chroot - specify chroot to do this in, False for host.
    """
    for package in packages.split():
        if chroot:
            os.system('chroot %s apt-get -d -y --reinstall install %s'
                      % (chroot, package))
        else:
            os.system('apt-get -d -y --reinstall install %s' % (package))


def recipe_run(config, globalconf):
    """
    Perform actions for apt module
    """
    if config['action'] == 'install':
        if config['download_only']:
            # The use case for download only is mostly for cases like live
            # media preperation where we want to set up a package pool for
            # the media
            download_only(globalconf['chroot'], config['packages'])
        else:
            install(config['chroot'], config['packages'])

    elif config['update-sources']:
        update(globalconf['chroot'])

    if config['action'] == 'clean':
        clean(globalconf['chroot'])
    if config['clean_cache']:
        clean(globalconf['chroot'])




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dmm-highvoltage",
    version="0.1.3",
    author="Jonathan Carter",
    author_email="jcc@debian.org",
    description="Distribution management modules",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://salsa.debian.org/highvoltage/dmm",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: ISC License (ISC)",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.6',
)




"""
file management functions for for dmm.
"""

import os
import pathlib
from command_runner import command_runner

# todo: add options to allow appending/replacing

def write_file(file, content):
    """
    Writes one or more lines to a file.
    """
    # Ensure that basepath exists first
    create_base_path(file)
    file = open(file, "w+")
    file.write(content)
    file.close()


def create_base_path(peth):
    """
    Ensures that a parent directory exists.
    """
    path = pathlib.Path(file).parent
    path.mkdir(parents=True, exist_ok=True)


def delete_path(path):
    print("Removing %s" % path)
    ecode, result = command_runner("rm -rf %s" % path)
    return ecode, result


def create_sparse_file(size, destination):
    """
    Create an empty sparse file.
    """
    ecode, result = command_runner("fallocate -l %s %s" % (size, destination))
    return ecode, result



"""
Generate checksums module for dmm.
"""

import os

def generate_checksums_file(sumbinary, path, output, cut):
    """
    Create a file that contains checksums:

    - sumbinary: md5sum, sha1sum, sha256sum
    - path: path that will be checksummed
    - output: where the checksums will be outputted to
              if empty, the result is returned
    - exclude: a list of lines that will include these
               strings will be excluded
    """
    checksum_output = os.popen("find %s -type f | xargs %s" % (path, sumbinary)).read()
    checksum_output_cut = checksum_output.replace(cut, '')
    if output == "":
        return (checksum_output_cut)
    else:
        file = open(output, "w+")
        file.write(checksum_output_cut)
        file.close()


def confirm_checksums(file):
    #TODO
    pass



"""
mksquashfs functions for for dmm.
"""

import os

def recipe_run(config, globalconf):
    """
    Perform actions for mksquashfs module
    """
    if config['action'] == 'make-squashfs':
        blocksize = config['blocksize']
        compress = config['compression-method']
        source = config['source']
        dest = config['destination']
        opts = config['options']
        os.system("mksquashfs %s %s -b %s -comp %s %s" % (source, dest, blocksize, compress, opts))

    #if config['action'] == unsquashfs:
        # todo




"""
readme functions for for dmm.
"""

import os
from datetime import date
# TODO: the date should come from our framework

def recipe_run(config, globalconf):
    """
    Perform actions for readme module
    """
    template = config['template']
    destination = config['destination'] + "/"
    filename = config['filename']
    product = config['name']
    version = config['version']
    readme = open(template).read().replace("__PRODUCT__", product).replace("__VERSION__", version).replace("__DATE__", str(date.today()))
    readme_w = open(destination + filename, 'w')
    readme_w.write(readme)
    readme_w.close()


